use std::path::Path;
use git2::Repository;

fn generate_callbacks(use_ssh_agent: bool, pb: &mut indicatif::ProgressBar) -> git2::RemoteCallbacks {
    let mut callbacks = git2::RemoteCallbacks::new();

    if use_ssh_agent {
        callbacks.credentials(|_, username, _| git2::Cred::ssh_key_from_agent(username.unwrap()));
    } else {
        callbacks.credentials(|_, username, _| git2::Cred::ssh_key(
            username.unwrap(),
            None,
            Path::new(&format!("{}/.ssh/id_rsa", std::env::var("HOME").unwrap_or(String::new()))),
            None
        ));
    }

    callbacks.transfer_progress(|progress| {
        if progress.total_deltas() == 0 {
            pb.set_length(progress.total_objects().max(1) as u64);
            pb.set_position(progress.received_objects() as u64);
            pb.set_message("Receiving objects");
        } else {
            pb.set_length(progress.total_deltas().max(1) as u64);
            pb.set_position(progress.total_objects() as u64);
            pb.set_message("Resolving deltas");
        }
        true
    });
    callbacks.push_transfer_progress(|cur, tot, _| {
        pb.set_length(tot.max(1) as u64);
        pb.set_position(cur as u64);
    });

    callbacks
}

fn _clone_repo(callbacks: git2::RemoteCallbacks) -> Result<Repository, git2::Error> {
    let mut fo = git2::FetchOptions::new();
    fo.remote_callbacks(callbacks);

    let mut builder = git2::build::RepoBuilder::new();
    builder.fetch_options(fo);

    builder.clone(
        "git@ssh.gitlab.mattv.de:root/dotfiles.git",
        Path::new(".")
    )
}

fn clone_repo() -> Result<Repository, git2::Error> {
    println!("Cloning repo...");
    let mut pb = indicatif::ProgressBar::new(1);
    pb.set_style(indicatif::ProgressStyle::with_template("{spinner} [{wide_bar}] {pos:>7}/{len:7} {msg}").unwrap().progress_chars("#>-"));
    _clone_repo(generate_callbacks(true, &mut pb))
        .or_else(|_| _clone_repo(generate_callbacks(false, &mut pb)))
}

fn _pull_repo(repo: &Repository, callbacks: git2::RemoteCallbacks) -> Result<(), git2::Error> {
    let mut fo = git2::FetchOptions::new();
    fo.remote_callbacks(callbacks);

    let mut co = git2::build::CheckoutBuilder::new();
    co.force();

    repo.find_remote("origin").unwrap().fetch(&["main"], Some(&mut fo), None)?;

    let mut commits = vec![];
    repo.fetchhead_foreach(|name, url, oid, was_merge| if was_merge { commits.push(repo.annotated_commit_from_fetchhead(name, &String::from_utf8_lossy(url), oid)); true} else { true })?;
    let commits = commits.into_iter().collect::<Result<Vec<_>, git2::Error>>()?;
    if commits.is_empty() {
        panic!("Failed to get heads from origin!");
    }

    repo.merge(commits.iter().collect::<Vec<_>>().as_slice(), None, Some(&mut co))
}

pub fn pull_repo(repo: &Repository) -> Result<(), git2::Error> {
    println!("Pulling repo...");
    let mut pb = indicatif::ProgressBar::new(1);
    pb.set_style(indicatif::ProgressStyle::with_template("{spinner} [{wide_bar}] {pos:>7}/{len:7} {msg}").unwrap().progress_chars("#>-"));
    _pull_repo(repo, generate_callbacks(true, &mut pb))
        .or_else(|_| _pull_repo(repo, generate_callbacks(false, &mut pb)))
}

fn _push_repo(repo: &Repository, callbacks: git2::RemoteCallbacks) -> Result<(), git2::Error> {
    let mut pu = git2::PushOptions::new();
    pu.remote_callbacks(callbacks);

    let mut remote = repo.find_remote("origin").unwrap();

    remote.push::<&str>(&["refs/heads/main:refs/heads/main"], Some(&mut pu))
}

pub fn push_repo(repo: &Repository) -> Result<(), git2::Error> {
    println!("Pushing repo...");
    let mut pb = indicatif::ProgressBar::new(1);
    pb.set_style(indicatif::ProgressStyle::with_template("{spinner} [{wide_bar}] {bytes}/{total_bytes} {msg}").unwrap().progress_chars("#>-"));
    _push_repo(repo, generate_callbacks(true, &mut pb))
        .or_else(|_| _push_repo(repo, generate_callbacks(false, &mut pb)))
}

pub fn open_repo() -> Result<Repository, String> {
    Repository::open(".")
        .or_else(|_| clone_repo())
        .map_err(|e| e.message().to_string())
}

pub fn is_clean(repo: &Repository) -> bool {
    let mut so = git2::StatusOptions::new();
    so.show(git2::StatusShow::IndexAndWorkdir);
    so.include_untracked(true);
    so.include_ignored(false);
    so.include_unmodified(false);

    repo.statuses(Some(&mut so)).map_or(false, |s| s.is_empty())
}

pub fn commit_changes(repo: &Repository, msg: String) {
    let sig = git2::Signature::now("Mutzi", "root@mattv.de").unwrap();

    let mut index = repo.index().unwrap();
    index.add_all(["*"], git2::IndexAddOption::DEFAULT, None).unwrap();

    let tree_id =  index.write_tree().unwrap();
    let tree = repo.find_tree(tree_id).unwrap();
    let parent_id = repo.refname_to_id("HEAD").unwrap();
    let parent = repo.find_commit(parent_id).unwrap();

    repo.commit(Some("HEAD"), &sig, &sig, msg.as_str(), &tree, &[&parent]).unwrap();
}
